# $Id: remote.tcl 721 2006-09-17 20:27:20Z sergei $
# Implementation of JEP-146 (Remote Controlling Clients) for Tkabber.
# http://www.jabber.org/jeps/jep-0146.html
# http://www.jabber.org/jeps/jep-0050.html

namespace eval ::remote {
    array set commands {}
    array set sessions {}

    custom::defgroup {Remote Control} \
	[::msgcat::mc "Remote control options"] -group Tkabber

    custom::defvar options(enable) 1 \
	[::msgcat::mc "Enable remote control."] \
	-type boolean -group {Remote Control}

    custom::defvar options(accept_from_myjid) 1 \
	[::msgcat::mc "Accept connections from my own JID."] \
	-type boolean -group {Remote Control}

    custom::defvar options(accept_list) "" \
	[::msgcat::mc "Accept connections from the listed JIDs."] \
	-type string -group {Remote Control}

    #custom::defvar options(show_my_resources) 1 \
    #	[::msgcat::mc "Show my own resources in the roster."] \
    #	-type boolean -group {Remote Control}
}

############################################

proc ::remote::allow_remote_control {connid from} {
    variable options

    if {!$options(enable)} {
	return 0
    }

    set from [string tolower $from]
    set myjid [string tolower \
		      [node_and_server_from_jid \
			   [jlib::connection_jid $connid]]]
    set bare_from [string tolower [node_and_server_from_jid $from]]

    if {$options(accept_from_myjid) && [cequal $myjid $bare_from]} {
	return 1
    }

    set accept_list [split [string tolower $options(accept_list)] " "]
    if {$bare_from != "" && [lsearch -exact $accept_list $bare_from] >= 0} {
	return 1
    }

    return 0
}

############################################
# Register and announce commands via disco

proc ::remote::register_command {node command name args} {
    variable commands
      
    set commands(command,$node) $command
    set commands(name,$node) $name
    lappend commands(nodes) $node

    ::disco::register_subnode $node \
	[namespace current]::common_command_infoitems_handler $name
}

proc ::remote::common_command_infoitems_handler {type connid from xmllist} {
    variable commands

    if {![allow_remote_control $connid $from]} {
	return {error auth forbidden}
    }

    jlib::wrapper:splitxml $xmllist tag vars isempty chdata children
    set node [jlib::wrapper:getattr $vars node]

    if {![cequal $node ""] && [info exists commands(command,$node)]} {
	if {[cequal $type info]} {
	    return \
		[list [jlib::wrapper:createtag identity \
			   -vars [list category automation \
				       type command-node \
				       name $commands(name,$node)]] \
		      [jlib::wrapper:createtag feature \
			   -vars [list var $::NS(commands)]]]
	} else {
	    return {}
	}
    } else { 
	return {error modify bad-request} 
    }
}

proc ::remote::commands_list_handler {type connid from xmllist} {
    variable commands

    if {![allow_remote_control $connid $from]} {
	return {error auth forbidden}
    }

    set myjid [jlib::connection_jid $connid]
    
    switch -- $type {
	items {
	    set items {}	    
	    foreach node $commands(nodes) {
		lappend items [jlib::wrapper:createtag item \
				   -vars [list jid $myjid \
					       node $node \
					       name $commands(name,$node)]]
	    }
	    return $items
	}
	info {
	    return [list [jlib::wrapper:createtag identity \
			      -vars [list category automation \
					  type command-list \
					  name "Remote control"]]]
	}
    }
    return {} 
}

::disco::register_feature $::NS(commands) {return ""}
::disco::register_node $::NS(commands) \
		       ::remote::commands_list_handler "Remote control"

#######################################
# Base engine.

proc ::remote::clear_session {sid node} {
    variable commands 
    variable sessions 

    if {![info exists commands(command,$node)]} return
    
    $commands(command,$node) $sid cancel {} 

    catch {unset sessions(node,$sid)}
    catch {unset sessions(connid,$sid)}
    catch {unset sessions(from,$sid)}	                
}

proc ::remote::create_session {node connid from} {
    variable commands    
    variable sessions
    
    if {![info exists commands(command,$node)]} return
    
    set cur_time [clock seconds]
    set rnd [random 1000]

    while {[info exists sessions(node,"$cur_time-$rnd")]} {
	set cur_time [clock seconds]
	set rnd [random 1000]
    }
    
    set sid "$cur_time-$rnd"

    set sessions(node,$sid) $node
    set sessions(connid,$sid) $connid
    set sessions(from,$sid) $from

    return $sid
}

proc ::remote::command_set_handler {connid from child} {   
    variable commands    
    variable sessions

    if {![allow_remote_control $connid $from]} {
	return {error auth forbidden}
    }

    jlib::wrapper:splitxml $child tag vars isempty chdata children

    set node [jlib::wrapper:getattr $vars node] 
    set action [jlib::wrapper:getattr $vars action] 
    set sid [jlib::wrapper:getattr $vars sessionid] 

    if {![info exists commands(command,$node)]} { 
	return {error cancel item-not-found}
    }
       
    if {[cequal $sid ""]} {
	set sid [create_session $node $connid $from]
    } else {
	if { ![info exists sessions(node,$sid)] } {
	    return [list error cancel bad-request \
			 -application-specific \
			     [jlib::wrapper:createtag session-expired \
				  -vars [list xmlns $::NS(commands)]]]
	}

	if {![cequal $sessions(node,$sid) $node] || \
	    ![cequal $sessions(connid,$sid) $connid] || \
	    ![cequal $sessions(from,$sid) $from]} {
	    return [list error cancel bad-request \
			 -application-specific \
			     [jlib::wrapper:createtag bad-sessionid \
				  -vars [list xmlns $::NS(commands)]]]
	}
    } 

    if {[cequal $action cancel]} {
	clear_session $sid $node
	return [list result [jlib::wrapper:createtag command \
				 -vars [list xmlns $::NS(commands) \
					     sessionid $sid \
					     node $node \
					     status canceled]]]
    }

    set result [$commands(command,$node) $sid $action $children]
    
    set status [lindex $result 0]
    switch -- $status {
	error { 
	    clear_session $sid $node
	    return $result
	}
	completed {
	    clear_session $sid $node
	}
	executing {}
	default {
	    clear_session $sid $node
	    return {error wait internal-server-error}
	}
    }            

    return [list result [jlib::wrapper:createtag command \
			     -vars [list xmlns $::NS(commands) \
					 sessionid $sid \
					 node $node \
					 status $status] \
			     -subtags [lrange $result 1 end]]]
} 
iq::register_handler set command $::NS(commands) ::remote::command_set_handler

# ##########################################
# Common functions for command implementations. 

# Scheduler for one-step dialogs and wizards
proc ::remote::standart_scheduler {steps prefix sid action children} {
    variable sessions
     
    if {[cequal $action cancel]} {
	catch { unset sessions($prefix,step,$sid) }
	for {set i 1} {$i <= $steps} {incr i} {
	    $prefix:clear_step$i $sid
	}
	return
    }

    if {![info exists sessions($prefix,step,$sid)] } {
	# First step

	if {![cequal $action execute] && ![cequal $action ""]} {
	    return [list error cancel bad-request \
			 -application-specific \
			     [jlib::wrapper:createtag bad-action \
				  -vars [list xmlns $::NS(commands)]]]
	}

	set sessions($prefix,step,$sid) 1
	return [$prefix:get_step$sessions($prefix,step,$sid) $sid]
	
    } elseif {$sessions($prefix,step,$sid) < $steps} {
	# Inner step

	if {![cequal $action execute] && ![cequal $action ""]} {
	    return [list error cancel bad-request \
			 -application-specific \
			     [jlib::wrapper:createtag bad-action \
				  -vars [list xmlns $::NS(commands)]]]
	}
	
	set res [$prefix:set_step$sessions($prefix,step,$sid) $sid $children]
	if {[cequal [lindex $res 0] error]} {
	    return $res
	}
	
	incr sessions($prefix,step,$sid) 	
	return [$prefix:get_step$sessions($prefix,step,$sid) $sid]
	
    } else {
	# Last step
	if {![cequal $action complete] && ![cequal $action ""]} {	    
	    return [list error cancel bad-request \
			 -application-specific \
			     [jlib::wrapper:createtag bad-action \
				  -vars [list xmlns $::NS(commands)]]]
	}
	
	set res [$prefix:set_step$sessions($prefix,step,$sid) $sid $children]
	if {[cequal [lindex $res 0] error]} {
	    return $res
	}
	
	return [$prefix:get_finish $sid]
    }
}

# Parses form result and returns array with values, check for correct form type
proc ::remote::standart_parseresult {children_b form_type} {
    set result {}
    foreach child $children_b {
	jlib::wrapper:splitxml $child tag vars isempty chdata children

	set xmlns  [jlib::wrapper:getattr $vars xmlns]
	set type  [jlib::wrapper:getattr $vars type]
	if {![cequal $tag x] || ![cequal $xmlns $::NS(data)]} {
	    continue
	}
	if {![cequal $type submit]} {
	    return [list error cancel bad-request \
			 -application-specific \
			     [jlib::wrapper:createtag bad-payload \
				  -vars [list xmlns $::NS(commands)]]]
	}
	
	foreach field [::data::parse_xdata_results $children -hidden 1] {
	    lassign $field var label value
	    if {[cequal $var FORM_TYPE]} {
		if {![cequal $value $form_type]} {
		    return [list error cancel bad-request \
				 -application-specific \
				     [jlib::wrapper:createtag bad-payload \
					  -vars [list xmlns $::NS(commands)]]] 
		}
	    } else {
		lappend result $var $value
	    }
	}
    }

    return $result
}

############################
#Change status
proc ::remote::change_status {sid action children} {
	return [standart_scheduler 1 change_status $sid $action $children]
}
::remote::register_command "http://jabber.org/protocol/rc#set-status" \
			   ::remote::change_status "Change status"

# step1: 
# send standart form
proc ::remote::change_status:get_step1 {sid} {
    global userstatus
    global textstatus
    global userpriority

    set fields {}

    lappend fields [data::createfieldtag hidden \
			-var FORM_TYPE \
			-values "http://jabber.org/protocol/rc"]

    lappend fields [data::createfieldtag title \
			-value "Change Status"]
    lappend fields [data::createfieldtag instructions \
			-value "Choose the status and status message"]
    
    set options {}
    foreach status {available chat away xa dnd unavailable} {
	lappend options [list $status $::statusdesc($status)]
    }
    lappend fields [data::createfieldtag list-single \
			-var status \
			-label "Status" \
			-required 1 \
			-value $userstatus \
			-options $options]
    lappend fields [data::createfieldtag text-single \
			-var status-priority \
			-label "Priority" \
			-value $userpriority \
			-required 1]
    lappend fields [data::createfieldtag text-multi \
			-var status-message \
			-label "Message" \
			-values [split $textstatus "\n"]]
    
    return [list executing [jlib::wrapper:createtag x \
				-vars [list xmlns $::NS(data) \
					    type form] \
				-subtags $fields]]

}

proc ::remote::change_status:set_step1 {sid children} {
    variable sessions
    
    set result [remote::standart_parseresult $children \
					     "http://jabber.org/protocol/rc"]

    if {[cequal [lindex $result 0] error]} {
	return $result
    }
    array set params $result
    
    if {![info exists params(status)] || \
	![info exists params(status-priority)]} {
	return [list error cancel bad-request \
		     -application-specific \
			 [jlib::wrapper:createtag bad-payload \
			      -vars [list xmlns $::NS(commands)]]]
    }
    
    set sessions(change_status,textstatus,$sid) {}
    catch {
	set sessions(change_status,textstatus,$sid) \
	    [join $params(status-message) "\n"]
    }
    set sessions(change_status,userstatus,$sid) \
	[lindex $params(status) 0]
    set sessions(change_status,userpriority,$sid) \
	[lindex $params(status-priority) 0]
    
    return {}
}

proc ::remote::change_status:clear_step1 {sid} {
    variable sessions 
    
    catch { unset sessions(change_status,textstatus,$sid) }
    catch { unset sessions(change_status,userstatus,$sid) }
    catch { unset sessions(change_status,userpriority,$sid) }
}

# finish:
# change status
# report
proc ::remote::change_status:get_finish {sid} {
    variable sessions
    global userstatus
    global textstatus 
    global userpriority 

    set textstatus $sessions(change_status,textstatus,$sid)
    set userpriority $sessions(change_status,userpriority,$sid)
    set userstatus $sessions(change_status,userstatus,$sid)

    return [list completed [jlib::wrapper:createtag note \
				-vars {type info} \
				-chdata "Status was changed successfully"]]
}


############################
# Leave groupchats
proc ::remote::leave_groupchats {sid action children} {
    return [standart_scheduler 1 leave_groupchats $sid $action $children]
}
::remote::register_command "http://jabber.org/protocol/rc#leave-groupchats" \
			   ::remote::leave_groupchats "Leave groupchats"

# step1: 
# allow users to choose which chats to leave
proc ::remote::leave_groupchats:get_step1 {sid} {
    variable sessions

    set options {}
    set connid $sessions(connid,$sid)
    foreach chatid [chat::opened] {
	set jid [chat::get_jid $chatid]

	if {![cequal [chat::get_connid $chatid] $connid]} continue
	if {![chat::is_groupchat $chatid]} continue
	if {[cequal [get_jid_presence_info show $connid $jid] ""]} continue

	set nick [get_our_groupchat_nick $chatid]

	lappend options [list $jid [format "%s at %s" $nick $jid]]
    }
    if {[expr [llength $options] == 0]} {
	return [list completed [jlib::wrapper:createtag note \
				    -vars {type info} \
				    -chdata "No groupchats to leave"]]
    }
    
    set fields {}

    lappend fields [data::createfieldtag hidden \
			-var FORM_TYPE \
			-values "http://jabber.org/protocol/rc"]
    lappend fields [data::createfieldtag title \
			-value "Leave Groupchats"]
    lappend fields [data::createfieldtag instructions \
			-value "Choose the groupchats you want to leave"]

    lappend fields [data::createfieldtag boolean \
			-var x-all \
			-label "Leave all groupchats" -value 0]
    lappend fields [data::createfieldtag list-multi \
			-var groupchats \
			-label "Groupchats" \
			-required 1 \
			-options $options]
    lappend fields [data::createfieldtag text-single \
			-var x-reason \
			-label "Reason"]
    
    return [list executing [jlib::wrapper:createtag x \
				-vars [list xmlns $::NS(data) \
					    type form] \
				-subtags $fields]]
}

proc ::remote::leave_groupchats:set_step1 {sid children} {
    variable sessions

    set result [remote::standart_parseresult $children \
					     "http://jabber.org/protocol/rc"]
    if {[cequal [lindex $result 0] error]} {
	return $result
    }
    array set params $result
    
    if {![info exists params(groupchats)]} {
	return [list error cancel bad-request \
		     -application-specific \
			 [jlib::wrapper:createtag bad-payload \
			      -vars [list xmlns $::NS(commands)]]]
    }
    
    set sessions(leave_groupchats,all,$sid) $params(all)
    set sessions(leave_groupchats,groupchats,$sid) $params(groupchats)
    set sessions(leave_groupchats,reason,$sid) ""
    catch {
	set sessions(leave_groupchats,reason,$sid) [lindex $params(reason) 0]
    }
    return {}

}

proc ::remote::leave_groupchats:clear_step1 {sid} {
    catch {unset sessions(leave_groupchats,groupchats,$sid)}
    catch {unset sessions(leave_groupchats,reason,$sid)}
    catch {unset sessions(leave_groupchats,all,$sid)}
}

# finish step 
# leave groupchats.
# report 
proc ::remote::leave_groupchats:get_finish {sid} {
    variable sessions

    set args [list  -connection $sessions(connid,$sid)]

    if {![cequal $sessions(leave_groupchats,reason,$sid) ""]} {
	lappend args -stat $sessions(leave_groupchats,reason,$sid)
    }

    # "all" workaround, will be removed soon
    if $sessions(leave_groupchats,all,$sid) {
	set connid $sessions(connid,$sid)
	set sessions(leave_groupchats,groupchats,$sid) ""
	
	foreach chatid [chat::opened] {
	    set jid [chat::get_jid $chatid]

	    if {![cequal [chat::get_connid $chatid] $connid]} continue
	    if {![chat::is_groupchat $chatid]} continue
	    if {[cequal [get_jid_presence_info show $connid $jid] ""]} continue
	    
	    lappend sessions(leave_groupchats,groupchats,$sid) $jid
	}
    }
    
    foreach jid $sessions(leave_groupchats,groupchats,$sid) {
	eval [list send_presence unavailable -to $jid] $args
    }

    return [list completed [jlib::wrapper:createtag note \
				-vars {type info} \
				-chdata "Groupchats was leaved successfully"]]
}

################################
# Forward unread messages

proc ::remote::forward {sid action children} {
    return [standart_scheduler 1 forward $sid $action $children]
}
::remote::register_command "http://jabber.org/protocol/rc#forward" \
			   ::remote::forward "Forward unread messages"

# step1: 
# form with list of unreaded correspondence
proc ::remote::forward:get_step1 {sid} {
    variable sessions

    set options {}
    set connid $sessions(connid,$sid)
    foreach id [array names ::remote::forward::unread] {
	lassign $id type chatid
	if {![cequal [chat::get_connid $chatid] $connid]} continue

	set jid [chat::get_jid $chatid]
	set name [::roster::itemconfig $connid \
				       [::roster::find_jid $connid $jid] \
				       -name]
	if {![cequal $name ""]} {
	    set name [format "%s (%s)" $name $jid]
	} else {
	    set name $jid
	}

	set count [llength $::remote::forward::unread($id)]

	lappend options [list $id \
			      [format "%s: %s %s message(s)" $name $count $type]]
    }
    if {[expr [llength $options] == 0]} {
	return [list completed [jlib::wrapper:createtag note \
				    -vars {type info} \
				    -chdata "There is no unread messages"]]
    }
    
    set fields {}

    lappend fields [data::createfieldtag hidden \
    			-var FORM_TYPE \
    			-values "tkabber:plugins:remote:forward_form"]
    lappend fields [data::createfieldtag title \
			-value "Forward Unread Messages"]
    lappend fields [data::createfieldtag instructions \
			-value "Choose from wich chats or groupchats you want to"]
    
    lappend fields [data::createfieldtag boolean \
			-var all \
			-label "Forward all messages" \
			-value 0]
    lappend fields [data::createfieldtag list-multi \
			-var chats -label "Forward messages from" \
			-required 1 \
			-options $options]
    
    return [list executing [jlib::wrapper:createtag x \
				-vars [list xmlns $::NS(data) \
					    type form] \
				-subtags $fields]]
}

proc ::remote::forward:set_step1 {sid children} {
    variable sessions
    
    set result [remote::standart_parseresult $children \
					     "plugins:remote:forward_form"]
    if {[cequal [lindex $result 0] error]} {
	return $result
    }
    array set params $result
    
    if {![info exists params(chats)]} {
	return [list error cancel bad-request \
		     -application-specific \
			 [jlib::wrapper:createtag bad-payload \
			      -vars [list xmlns $::NS(commands)]]]
    }

    set sessions(forward,all,$sid) $params(all)
    set sessions(forward,chats,$sid) $params(chats)
    return {}
}

proc ::remote::forward:clear_step1 {sid} {
    catch {unset sessions(forward,chats,$sid)}
    catch {unset sessions(forward,all,$sid)}
}

# finish: 
# forward selected unread messages
# report
proc ::remote::forward:get_finish {sid} {
    variable sessions

    set connid $sessions(connid,$sid)
    set oto [jlib::connection_jid $connid]
    set target $sessions(from,$sid)
    
    # "all" workaround, will be removed soon
    if $sessions(forward,all,$sid) {
	set sessions(forward,chats,$sid) {}
	
	foreach id [array names forward::unread] {
	    lassign $id type chatid	    
	    if {![cequal [chat::get_connid $chatid] $connid]} continue
	    lappend sessions(forward,chats,$sid) $id
	}
    }

    foreach id $sessions(forward,chats,$sid) {
	forward::forward_messages $id $connid $oto $target
    }

    return [list completed \
		 [jlib::wrapper:createtag note \
		      -vars {type info} \
		      -chdata "Unread messages was forwarded successfully"]]
}

#############################
# Forward namespace
namespace eval ::remote::forward {
    array set unread {}
}

# forwards messages
# leaves marks that they was forwarded. 
# cleanup arrays
proc ::remote::forward::forward_messages {id connid oto target} {
    variable unread
    variable msgdata

    lassign $id type chatid
    
    foreach elem $unread($id) {
	
	switch -- $type {
	    groupchat -
	    chat {
		lassign $elem date ofrom body x
	    }
	    normal {
		lassign $msgdata($elem) date ofrom body x
	    }
	}
	
	lappend x [jlib::wrapper:createtag addresses \
		       -vars [list xmlns $::NS(xaddress)] \
		       -subtags [list [jlib::wrapper:createtag address \
					   -vars [list type ofrom \
						       jid $ofrom]] \
				      [jlib::wrapper:createtag address \
					   -vars [list type oto \
						       jid $oto]]]]
	
	lappend x [jlib::wrapper:createtag x \
		       -vars [list xmlns "jabber:x:delay" \
				   stamp $date]]
	
	jlib::send_msg $target -body $body \
			       -type $type \
			       -xlist $x \
			       -connection $connid

	switch -- $type {
	    normal {
		set lab \
		    [Label $elem.forwlab \
			   -text [::msgcat::mc \
				      "This message was forwarded to %s" \
				      $target]]
		pack $lab -anchor w -fill none -expand no -before $elem.title
		
		catch {unset msgdata($elem)}
	    }
	}
    }

    catch {unset unread($id)}
    switch -- $type {
	groupchat -
	chat {   
	    after idle \
		  [list ::chat::add_message $chatid $ofrom info \
			[::msgcat::mc "All unread messages was forwarded to %s." \
			     $target] \
			{}]
	}
    }
}

# store message into the unread if type == chat
proc ::remote::forward::draw_message_handler {chatid from type body extras} {
    variable unread
    
    if {[ifacetk::chat_window_is_active $chatid]} return 

    if {![lcontain {chat groupchat} $type]} return
#    if {![cequal chat $type]} return

    set date [clock format [clock seconds] -format "%Y%m%dT%H:%M:%S" -gmt 1]

    set message [list $date $from $body $extras]
    set id [list $type $chatid]
    lappend unread($id) $message

    return 0
}

hook::add draw_message_hook ::remote::forward::draw_message_handler 19

# clear list of unread messages with type == chat
proc ::remote::forward::trace_number_msg {var1 chatid mode} {
    variable unread
    
    if { $::ifacetk::number_msg($chatid) == 0 } {
	set type $::chat::chats(type,$chatid)
	set id [list $type $chatid]
	catch {unset unread($id)}
    }
    
}

trace variable ::ifacetk::number_msg r ::remote::forward::trace_number_msg

# store message with type == normal
proc ::remote::forward::message_process_x \
     {rowvar bodyvar f x connid from type replyP} {
    upvar 2 $rowvar row
    upvar 2 $bodyvar body
    variable unread
    variable msgdata
   
    if {!$replyP || [cequal $type error]} {
	return
    }
    
    set id [list normal [chat::chatid $connid $from]]

    if {![info exists unread($id)]} {
	set unread($id) {}
    }

    set msgwin [winfo toplevel $f]

    bindtags $msgwin [concat [bindtags $msgwin] tag$msgwin]
    bind tag$msgwin <Destroy> \
	 +[list [namespace current]::on_msgwin_destroy $msgwin $id]
    lappend unread($id) $msgwin

    set date [clock format [clock seconds] -format "%Y%m%dT%H:%M:%S" -gmt 1]
    set msgdata($msgwin) [list $date $from $body $x]

    return
}

hook::add message_process_x_hook ::remote::forward::message_process_x

# clear chat message with type == normal if it was closed
proc ::remote::forward::on_msgwin_destroy {msgwin id} {
    variable unread
    variable msgdata
    
    if {![info exists unread($id)]} return

    if {[set index [lsearch -exact $unread($id) $msgwin]] >= 0} {
	set unread($id) [lreplace $unread($id) $index $index]
	catch {unset msgdata($msgwin)}
    }    

    if { [llength $unread($id)] == 0 } {
	catch {unset unread($id)}
    }
}

